import ast
import inspect as inspect_ast
import os
from datetime import datetime
from pathlib import Path

from alembic import command
from alembic.config import Config as AlembicConfig
from alembic.script import Script, ScriptDirectory
from omni.pro import redis
from omni.pro.config import Config
from omni.pro.logger import configure_logger
from sqlalchemy import Column, String, create_engine
from sqlalchemy import inspect as inspect_sqlalchemy
from sqlalchemy.orm import DeclarativeBase, sessionmaker

logger = configure_logger(name=__name__)


class Base(DeclarativeBase):
    pass


class AlembicVersionModel(Base):
    __tablename__ = "alembic_version"

    version_num = Column(String(32), primary_key=True)


class AlembicMigrateCheck(object):
    def __init__(self, path: Path, postgres_url: str = None) -> None:
        self.redis_manager = redis.get_redis_manager()
        self.tenants = self.redis_manager.get_tenant_codes()
        self.alembic_config = AlembicConfig(path / "alembic.ini")
        self.alembic_config.set_main_option("script_location", str(path / "alembic"))
        self.versions_path = path / "alembic" / "versions"
        self.postgres_url = postgres_url
        self.changes = False

    def main(self):
        logger.info("Start validate")
        last_version = None
        validations = []
        for idx, tenant in enumerate(self.tenants):
            logger.info(f"Migrate tenant: {tenant}")
            # try:
            host, port, user, password, name = self.get_postgres_config(Config.SERVICE_ID, tenant)
            if not all([host, port, user, password, name]):
                raise ValueError(f"Invalid database config for tenant: {tenant}")
            sql_connect = f"postgresql://{user}:{password}@{host}:{port}/{name}"
            self.postgres_url = sql_connect
            if idx == 0:
                last_version = self.apply()
            else:
                self.upgrade_head(last_version)
            # except Exception as e:
            #     logger.error(f"Failed to migrate tenant {tenant}: {e}")

        if self.changes:
            print(f"REPO_URL={self.redis_manager.get_json(f'SETTINGS', f'repos.{Config.SERVICE_ID}.url')}")
            print(f"RUN=1")

    @property
    def postgres_url(self) -> str:
        return self._postgres_url

    @postgres_url.setter
    def postgres_url(self, postgres_url) -> None:
        self._postgres_url = postgres_url
        if postgres_url:
            self.alembic_config.set_main_option("sqlalchemy.url", postgres_url)
            self.engine = create_engine(postgres_url)
            self.Session = sessionmaker(bind=self.engine)

    def get_postgres_config(self, service_id, tenant_code):
        postres_config = self.redis_manager.get_postgres_config(Config.SERVICE_ID, tenant_code)
        host, port, user, password, name = postres_config.values()
        if all([host, port, user, password, name]):
            return host, port, user, password, name
        logger.error(f"Postgres config not found for tenant: {tenant_code}")

    def check(self) -> bool:
        # Comprobar si la tabla alembic_version existe
        inspector = inspect_sqlalchemy(self.engine)
        if not inspector.has_table(AlembicVersionModel.__tablename__):
            AlembicVersionModel.metadata.create_all(self.engine)
            return False
        return True

    def get_current_version(self) -> str:
        with self.Session() as session:
            latest_version = session.query(AlembicVersionModel).order_by(AlembicVersionModel.version_num.desc()).first()
            return latest_version.version_num if latest_version else None

    def apply_revision(self, message="auto-generated revision") -> Script:
        # Usa la librería alembic para aplicar la revisión
        # dar formato al mensaje concatenando la fecha y hora en formato ISO
        message = f"{message} {datetime.now().isoformat()}"
        new_revision = command.revision(self.alembic_config, autogenerate=True, message=message)
        return new_revision

    def no_changes_detected(self, script: Script) -> bool:
        code = inspect_ast.getsource(script.module.upgrade)
        tree = ast.parse(code)
        clean_code = ast.unparse(tree)

        lines = clean_code.split("\n")
        l = [line for line in lines if line.strip() not in ["", "def upgrade():", "def upgrade() -> None:", "pass"]]
        return not l

    def upgrade_head(self, revision=None) -> None:
        # Usa la librería alembic para hacer las migraciones
        command.upgrade(self.alembic_config, revision or "head")

    def apply(self):
        # Uso del AlembicMigrateCheck
        self.check()
        current_version = self.get_current_version()
        script_directory = ScriptDirectory.from_config(self.alembic_config)
        # is last_version es None es la primera ves que se crea la tabla alembic_version
        if current_version is None and list(script_directory.walk_revisions()):
            self.upgrade_head()
            current_version = self.get_current_version()
        script = self.apply_revision()
        if self.no_changes_detected(script):
            self.last_version = current_version
            Path(script.path).unlink()
        else:
            self.last_version = script.revision
            self.changes = True

        self.upgrade_head(self.last_version)
        return self.last_version



import boto3


class AWSClient(object):
    def __init__(
        self, service_name: str, region_name: str, aws_access_key_id: str, aws_secret_access_key: str, **kwargs
    ) -> None:
        """
        :type service_name: str
        :param service_name: AWS service name
        :type region_name: str
        :param region_name: AWS region name
        :type aws_access_key_id: str
        :param aws_access_key_id: AWS access key id
        :type aws_secret_access_key: str
        :param aws_secret_access_key: AWS secret access key
        Example:
            service_name = "service_name"
            region_name = "us-east-1"
            aws_access_key_id = "AKIAIOSFODNN7EXAMPLE"
            aws_secret_access_key = "wJalrXUtnFEMI/K7MDENG/bPxRfiCYEXAMPLEKEY"
        """
        self._client = boto3.client(
            service_name=service_name,
            region_name=region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            **kwargs,
        )

    def get_client(self):
        return self._client

    client = property(get_client)



from omni.pro.aws.client import AWSClient


class AWSCloudMap(AWSClient):
    def __init__(
        self,
        region_name: str,
        namespace_name: str,
        service_name: str,
        aws_access_key_id: str,
        aws_secret_access_key: str,
        **kwargs,
    ):
        super().__init__(
            service_name="servicediscovery",
            region_name=region_name,
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            **kwargs,
        )
        self.namespace_name = namespace_name
        self.service_name = service_name

    def discover_instances(self):
        response = self.client.discover_instances(
            NamespaceName=self.namespace_name,
            ServiceName=self.service_name,
        )
        if not response.get("Instances"):
            raise Exception("No instances found")

        return response.get("Instances")

    def get_redis_config(self):
        instances = self.discover_instances()
        instance = instances[0]
        return {
            "host": instance.get("Attributes").get("host"),
            "port": int(instance.get("Attributes").get("port") or 6379),
            "db": int(instance.get("Attributes").get("db") or 0),
        }



from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from omni.pro.protos.common import base_pb2 as _base_pb2

DESCRIPTOR: _descriptor.FileDescriptor

class RedistributionOrder(_message.Message):
    __slots__ = ["id", "distribution_data", "object_audit"]
    ID_FIELD_NUMBER: _ClassVar[int]
    DISTRIBUTION_DATA_FIELD_NUMBER: _ClassVar[int]
    OBJECT_AUDIT_FIELD_NUMBER: _ClassVar[int]
    id: str
    distribution_data: _struct_pb2.Struct
    object_audit: _base_pb2.ObjectAudit
    def __init__(
        self,
        id: _Optional[str] = ...,
        distribution_data: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...,
        object_audit: _Optional[_Union[_base_pb2.ObjectAudit, _Mapping]] = ...,
    ) -> None: ...

class RedistributionOrderCreateRequest(_message.Message):
    __slots__ = ["action", "context"]
    ACTION_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    action: bool
    context: _base_pb2.Context
    def __init__(self, action: bool = ..., context: _Optional[_Union[_base_pb2.Context, _Mapping]] = ...) -> None: ...

class RedistributionOrderCreateResponse(_message.Message):
    __slots__ = ["message", "redistribution_order", "response_standard"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    REDISTRIBUTION_ORDER_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_STANDARD_FIELD_NUMBER: _ClassVar[int]
    message: str
    redistribution_order: RedistributionOrder
    response_standard: _base_pb2.ResponseStandard
    def __init__(
        self,
        message: _Optional[str] = ...,
        redistribution_order: _Optional[_Union[RedistributionOrder, _Mapping]] = ...,
        response_standard: _Optional[_Union[_base_pb2.ResponseStandard, _Mapping]] = ...,
    ) -> None: ...

class RedistributionOrderReadRequest(_message.Message):
    __slots__ = ["group_by", "sort_by", "fields", "filter", "paginated", "id", "context"]
    GROUP_BY_FIELD_NUMBER: _ClassVar[int]
    SORT_BY_FIELD_NUMBER: _ClassVar[int]
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    PAGINATED_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    group_by: _containers.RepeatedCompositeFieldContainer[_base_pb2.GroupBy]
    sort_by: _base_pb2.SortBy
    fields: _base_pb2.Fields
    filter: _base_pb2.Filter
    paginated: _base_pb2.Paginated
    id: str
    context: _base_pb2.Context
    def __init__(
        self,
        group_by: _Optional[_Iterable[_Union[_base_pb2.GroupBy, _Mapping]]] = ...,
        sort_by: _Optional[_Union[_base_pb2.SortBy, _Mapping]] = ...,
        fields: _Optional[_Union[_base_pb2.Fields, _Mapping]] = ...,
        filter: _Optional[_Union[_base_pb2.Filter, _Mapping]] = ...,
        paginated: _Optional[_Union[_base_pb2.Paginated, _Mapping]] = ...,
        id: _Optional[str] = ...,
        context: _Optional[_Union[_base_pb2.Context, _Mapping]] = ...,
    ) -> None: ...

class RedistributionOrderReadResponse(_message.Message):
    __slots__ = ["redistribution_orders", "response_standard", "meta_data"]
    REDISTRIBUTION_ORDERS_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_STANDARD_FIELD_NUMBER: _ClassVar[int]
    META_DATA_FIELD_NUMBER: _ClassVar[int]
    redistribution_orders: _containers.RepeatedCompositeFieldContainer[RedistributionOrder]
    response_standard: _base_pb2.ResponseStandard
    meta_data: _base_pb2.MetaData
    def __init__(
        self,
        redistribution_orders: _Optional[_Iterable[_Union[RedistributionOrder, _Mapping]]] = ...,
        response_standard: _Optional[_Union[_base_pb2.ResponseStandard, _Mapping]] = ...,
        meta_data: _Optional[_Union[_base_pb2.MetaData, _Mapping]] = ...,
    ) -> None: ...

class RedistributionOrderDeleteRequest(_message.Message):
    __slots__ = ["id", "context"]
    ID_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    id: str
    context: _base_pb2.Context
    def __init__(
        self, id: _Optional[str] = ..., context: _Optional[_Union[_base_pb2.Context, _Mapping]] = ...
    ) -> None: ...

class RedistributionOrderDeleteResponse(_message.Message):
    __slots__ = ["message", "response_standard"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_STANDARD_FIELD_NUMBER: _ClassVar[int]
    message: str
    response_standard: _base_pb2.ResponseStandard
    def __init__(
        self,
        message: _Optional[str] = ...,
        response_standard: _Optional[_Union[_base_pb2.ResponseStandard, _Mapping]] = ...,
    ) -> None: ...



from typing import Optional

from sqlalchemy.orm import mapped_column as mapped_column_alchemy
from sqlalchemy.orm.properties import MappedColumn


class MappedColumnOmni(MappedColumn):
    def __init__(self, *args, is_filterable=True, is_exportable=True, is_importable=True, **kwargs):
        super().__init__(*args, **kwargs)
        self.column.is_filterable = is_filterable
        self.column.is_exportable = is_exportable
        self.column.is_importable = is_importable


def mapped_column(
    *args,
    is_filterable: Optional[bool] = True,
    is_exportable: Optional[bool] = True,
    is_importable: Optional[bool] = True,
    **kwargs
):
    mc = mapped_column_alchemy(*args, **kwargs)
    mc.column.is_filterable = is_filterable
    mc.column.is_exportable = is_exportable
    mc.column.is_importable = is_importable
    return mc



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v1/clients/type_document.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from omni.pro.protos.common import base_pb2 as common_dot_base__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x1ev1/clients/type_document.proto\x12)pro.omni.oms.api.v1.clients.type_document\x1a\x11\x63ommon/base.proto\x1a\x1egoogle/protobuf/wrappers.proto"\xe1\x01\n\x0cTypeDocument\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0c\n\x04\x63ode\x18\x02 \x01(\t\x12*\n\x06is_vat\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x1c\n\x14\x64ocument_type_doc_id\x18\x04 \x01(\t\x12*\n\x06\x61\x63tive\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12?\n\x0cobject_audit\x18\x06 \x01(\x0b\x32).pro.omni.oms.api.common.base.ObjectAudit"\xb9\x01\n\x19TypeDocumentCreateRequest\x12\x0c\n\x04\x63ode\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x1c\n\x14\x64ocument_type_doc_id\x18\x03 \x01(\t\x12*\n\x06is_vat\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x36\n\x07\x63ontext\x18\x05 \x01(\x0b\x32%.pro.omni.oms.api.common.base.Context"\xb7\x01\n\x1aTypeDocumentCreateResponse\x12N\n\rtype_document\x18\x01 \x01(\x0b\x32\x37.pro.omni.oms.api.v1.clients.type_document.TypeDocument\x12I\n\x11response_standard\x18\x02 \x01(\x0b\x32..pro.omni.oms.api.common.base.ResponseStandard"\xf5\x02\n\x17TypeDocumentReadRequest\x12\x37\n\x08group_by\x18\x01 \x03(\x0b\x32%.pro.omni.oms.api.common.base.GroupBy\x12\x35\n\x07sort_by\x18\x02 \x01(\x0b\x32$.pro.omni.oms.api.common.base.SortBy\x12\x34\n\x06\x66ields\x18\x03 \x01(\x0b\x32$.pro.omni.oms.api.common.base.Fields\x12\x34\n\x06\x66ilter\x18\x04 \x01(\x0b\x32$.pro.omni.oms.api.common.base.Filter\x12:\n\tpaginated\x18\x05 \x01(\x0b\x32\'.pro.omni.oms.api.common.base.Paginated\x12\n\n\x02id\x18\x06 \x01(\t\x12\x36\n\x07\x63ontext\x18\x07 \x01(\x0b\x32%.pro.omni.oms.api.common.base.Context"\xf1\x01\n\x18TypeDocumentReadResponse\x12O\n\x0etype_documents\x18\x01 \x03(\x0b\x32\x37.pro.omni.oms.api.v1.clients.type_document.TypeDocument\x12I\n\x11response_standard\x18\x02 \x01(\x0b\x32..pro.omni.oms.api.common.base.ResponseStandard\x12\x39\n\tmeta_data\x18\x03 \x01(\x0b\x32&.pro.omni.oms.api.common.base.MetaData"\xa3\x01\n\x19TypeDocumentUpdateRequest\x12N\n\rtype_document\x18\x01 \x01(\x0b\x32\x37.pro.omni.oms.api.v1.clients.type_document.TypeDocument\x12\x36\n\x07\x63ontext\x18\x02 \x01(\x0b\x32%.pro.omni.oms.api.common.base.Context"\xb7\x01\n\x1aTypeDocumentUpdateResponse\x12N\n\rtype_document\x18\x01 \x01(\x0b\x32\x37.pro.omni.oms.api.v1.clients.type_document.TypeDocument\x12I\n\x11response_standard\x18\x02 \x01(\x0b\x32..pro.omni.oms.api.common.base.ResponseStandard"_\n\x19TypeDocumentDeleteRequest\x12\n\n\x02id\x18\x01 \x01(\t\x12\x36\n\x07\x63ontext\x18\x02 \x01(\x0b\x32%.pro.omni.oms.api.common.base.Context"g\n\x1aTypeDocumentDeleteResponse\x12I\n\x11response_standard\x18\x01 \x01(\x0b\x32..pro.omni.oms.api.common.base.ResponseStandard2\xa7\x05\n\x13TypeDocumentService\x12\xa3\x01\n\x12TypeDocumentCreate\x12\x44.pro.omni.oms.api.v1.clients.type_document.TypeDocumentCreateRequest\x1a\x45.pro.omni.oms.api.v1.clients.type_document.TypeDocumentCreateResponse"\x00\x12\x9d\x01\n\x10TypeDocumentRead\x12\x42.pro.omni.oms.api.v1.clients.type_document.TypeDocumentReadRequest\x1a\x43.pro.omni.oms.api.v1.clients.type_document.TypeDocumentReadResponse"\x00\x12\xa3\x01\n\x12TypeDocumentUpdate\x12\x44.pro.omni.oms.api.v1.clients.type_document.TypeDocumentUpdateRequest\x1a\x45.pro.omni.oms.api.v1.clients.type_document.TypeDocumentUpdateResponse"\x00\x12\xa3\x01\n\x12TypeDocumentDelete\x12\x44.pro.omni.oms.api.v1.clients.type_document.TypeDocumentDeleteRequest\x1a\x45.pro.omni.oms.api.v1.clients.type_document.TypeDocumentDeleteResponse"\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "v1.clients.type_document_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _globals["_TYPEDOCUMENT"]._serialized_start = 129
    _globals["_TYPEDOCUMENT"]._serialized_end = 354
    _globals["_TYPEDOCUMENTCREATEREQUEST"]._serialized_start = 357
    _globals["_TYPEDOCUMENTCREATEREQUEST"]._serialized_end = 542
    _globals["_TYPEDOCUMENTCREATERESPONSE"]._serialized_start = 545
    _globals["_TYPEDOCUMENTCREATERESPONSE"]._serialized_end = 728
    _globals["_TYPEDOCUMENTREADREQUEST"]._serialized_start = 731
    _globals["_TYPEDOCUMENTREADREQUEST"]._serialized_end = 1104
    _globals["_TYPEDOCUMENTREADRESPONSE"]._serialized_start = 1107
    _globals["_TYPEDOCUMENTREADRESPONSE"]._serialized_end = 1348
    _globals["_TYPEDOCUMENTUPDATEREQUEST"]._serialized_start = 1351
    _globals["_TYPEDOCUMENTUPDATEREQUEST"]._serialized_end = 1514
    _globals["_TYPEDOCUMENTUPDATERESPONSE"]._serialized_start = 1517
    _globals["_TYPEDOCUMENTUPDATERESPONSE"]._serialized_end = 1700
    _globals["_TYPEDOCUMENTDELETEREQUEST"]._serialized_start = 1702
    _globals["_TYPEDOCUMENTDELETEREQUEST"]._serialized_end = 1797
    _globals["_TYPEDOCUMENTDELETERESPONSE"]._serialized_start = 1799
    _globals["_TYPEDOCUMENTDELETERESPONSE"]._serialized_end = 1902
    _globals["_TYPEDOCUMENTSERVICE"]._serialized_start = 1905
    _globals["_TYPEDOCUMENTSERVICE"]._serialized_end = 2584
# @@protoc_insertion_point(module_scope)



from typing import ClassVar as _ClassVar
from typing import Iterable as _Iterable
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from google.protobuf.internal import containers as _containers
from omni.pro.protos.common import base_pb2 as _base_pb2
from omni.pro.protos.v1.users import user_pb2 as _user_pb2

DESCRIPTOR: _descriptor.FileDescriptor

class LogSession(_message.Message):
    __slots__ = ["id", "user", "operation", "date_start", "date_end", "object_audit"]
    ID_FIELD_NUMBER: _ClassVar[int]
    USER_FIELD_NUMBER: _ClassVar[int]
    OPERATION_FIELD_NUMBER: _ClassVar[int]
    DATE_START_FIELD_NUMBER: _ClassVar[int]
    DATE_END_FIELD_NUMBER: _ClassVar[int]
    OBJECT_AUDIT_FIELD_NUMBER: _ClassVar[int]
    id: str
    user: _user_pb2.User
    operation: str
    date_start: _timestamp_pb2.Timestamp
    date_end: _timestamp_pb2.Timestamp
    object_audit: _base_pb2.ObjectAudit
    def __init__(
        self,
        id: _Optional[str] = ...,
        user: _Optional[_Union[_user_pb2.User, _Mapping]] = ...,
        operation: _Optional[str] = ...,
        date_start: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        date_end: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ...,
        object_audit: _Optional[_Union[_base_pb2.ObjectAudit, _Mapping]] = ...,
    ) -> None: ...

class LogSessionReadRequest(_message.Message):
    __slots__ = ["group_by", "sort_by", "fields", "filter", "paginated", "id", "context"]
    GROUP_BY_FIELD_NUMBER: _ClassVar[int]
    SORT_BY_FIELD_NUMBER: _ClassVar[int]
    FIELDS_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    PAGINATED_FIELD_NUMBER: _ClassVar[int]
    ID_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    group_by: _containers.RepeatedCompositeFieldContainer[_base_pb2.GroupBy]
    sort_by: _base_pb2.SortBy
    fields: _base_pb2.Fields
    filter: _base_pb2.Filter
    paginated: _base_pb2.Paginated
    id: str
    context: _base_pb2.Context
    def __init__(
        self,
        group_by: _Optional[_Iterable[_Union[_base_pb2.GroupBy, _Mapping]]] = ...,
        sort_by: _Optional[_Union[_base_pb2.SortBy, _Mapping]] = ...,
        fields: _Optional[_Union[_base_pb2.Fields, _Mapping]] = ...,
        filter: _Optional[_Union[_base_pb2.Filter, _Mapping]] = ...,
        paginated: _Optional[_Union[_base_pb2.Paginated, _Mapping]] = ...,
        id: _Optional[str] = ...,
        context: _Optional[_Union[_base_pb2.Context, _Mapping]] = ...,
    ) -> None: ...

class LogSessionReadResponse(_message.Message):
    __slots__ = ["response_standard", "log_sessions", "meta_data"]
    RESPONSE_STANDARD_FIELD_NUMBER: _ClassVar[int]
    LOG_SESSIONS_FIELD_NUMBER: _ClassVar[int]
    META_DATA_FIELD_NUMBER: _ClassVar[int]
    response_standard: _base_pb2.ResponseStandard
    log_sessions: _containers.RepeatedCompositeFieldContainer[LogSession]
    meta_data: _base_pb2.MetaData
    def __init__(
        self,
        response_standard: _Optional[_Union[_base_pb2.ResponseStandard, _Mapping]] = ...,
        log_sessions: _Optional[_Iterable[_Union[LogSession, _Mapping]]] = ...,
        meta_data: _Optional[_Union[_base_pb2.MetaData, _Mapping]] = ...,
    ) -> None: ...

class PausedMethodRequest(_message.Message):
    __slots__ = ["id", "type", "context"]
    ID_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CONTEXT_FIELD_NUMBER: _ClassVar[int]
    id: str
    type: str
    context: _base_pb2.Context
    def __init__(
        self,
        id: _Optional[str] = ...,
        type: _Optional[str] = ...,
        context: _Optional[_Union[_base_pb2.Context, _Mapping]] = ...,
    ) -> None: ...

class PausedMethodResponse(_message.Message):
    __slots__ = ["response_standard", "log_session", "meta_data"]
    RESPONSE_STANDARD_FIELD_NUMBER: _ClassVar[int]
    LOG_SESSION_FIELD_NUMBER: _ClassVar[int]
    META_DATA_FIELD_NUMBER: _ClassVar[int]
    response_standard: _base_pb2.ResponseStandard
    log_session: LogSession
    meta_data: _base_pb2.MetaData
    def __init__(
        self,
        response_standard: _Optional[_Union[_base_pb2.ResponseStandard, _Mapping]] = ...,
        log_session: _Optional[_Union[LogSession, _Mapping]] = ...,
        meta_data: _Optional[_Union[_base_pb2.MetaData, _Mapping]] = ...,
    ) -> None: ...



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
from omni.pro.protos.v1.rules import appointment_pb2 as v1_dot_rules_dot_appointment__pb2


class AppointmentServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.AppointmentCreate = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentCreate",
            request_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentCreateRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentCreateResponse.FromString,
        )
        self.AppointmentRead = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentRead",
            request_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentReadRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentReadResponse.FromString,
        )
        self.AppointmentUpdate = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentUpdate",
            request_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentUpdateRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentUpdateResponse.FromString,
        )
        self.AppointmentDelete = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentDelete",
            request_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentDeleteRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentDeleteResponse.FromString,
        )
        self.SearchAppointment = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/SearchAppointment",
            request_serializer=v1_dot_rules_dot_appointment__pb2.SearchAppointmentRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.SearchAppointmentResponse.FromString,
        )
        self.ConfirmAppointment = channel.unary_unary(
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/ConfirmAppointment",
            request_serializer=v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentRequest.SerializeToString,
            response_deserializer=v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentResponse.FromString,
        )


class AppointmentServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def AppointmentCreate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AppointmentRead(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AppointmentUpdate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def AppointmentDelete(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def SearchAppointment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ConfirmAppointment(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_AppointmentServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "AppointmentCreate": grpc.unary_unary_rpc_method_handler(
            servicer.AppointmentCreate,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentCreateRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentCreateResponse.SerializeToString,
        ),
        "AppointmentRead": grpc.unary_unary_rpc_method_handler(
            servicer.AppointmentRead,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentReadRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentReadResponse.SerializeToString,
        ),
        "AppointmentUpdate": grpc.unary_unary_rpc_method_handler(
            servicer.AppointmentUpdate,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentUpdateRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentUpdateResponse.SerializeToString,
        ),
        "AppointmentDelete": grpc.unary_unary_rpc_method_handler(
            servicer.AppointmentDelete,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.AppointmentDeleteRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.AppointmentDeleteResponse.SerializeToString,
        ),
        "SearchAppointment": grpc.unary_unary_rpc_method_handler(
            servicer.SearchAppointment,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.SearchAppointmentRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.SearchAppointmentResponse.SerializeToString,
        ),
        "ConfirmAppointment": grpc.unary_unary_rpc_method_handler(
            servicer.ConfirmAppointment,
            request_deserializer=v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentRequest.FromString,
            response_serializer=v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "pro.omni.oms.api.v1.rules.appointment.AppointmentService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class AppointmentService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def AppointmentCreate(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentCreate",
            v1_dot_rules_dot_appointment__pb2.AppointmentCreateRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.AppointmentCreateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AppointmentRead(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentRead",
            v1_dot_rules_dot_appointment__pb2.AppointmentReadRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.AppointmentReadResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AppointmentUpdate(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentUpdate",
            v1_dot_rules_dot_appointment__pb2.AppointmentUpdateRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.AppointmentUpdateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def AppointmentDelete(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/AppointmentDelete",
            v1_dot_rules_dot_appointment__pb2.AppointmentDeleteRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.AppointmentDeleteResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def SearchAppointment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/SearchAppointment",
            v1_dot_rules_dot_appointment__pb2.SearchAppointmentRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.SearchAppointmentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ConfirmAppointment(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/pro.omni.oms.api.v1.rules.appointment.AppointmentService/ConfirmAppointment",
            v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentRequest.SerializeToString,
            v1_dot_rules_dot_appointment__pb2.ConfirmAppointmentResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import ast

import mongoengine as mongo
from bson import ObjectId
from omni.pro.protos.common import base_pb2
from pymongo import UpdateOne


class DatabaseManager(object):
    def __init__(self, host: str, port: int, db: str, user: str, password: str, complement: dict) -> None:
        """
        :param db_object: Database object
        Example:
            db_object = {
                "host":"mongo",
                "port":"27017",
                "user":"root",
                "password":"123456",
                "type":"write | read",
                "no_sql":"true",
                "complement":""
            }
        """
        self.db = db
        self.host = host
        self.port = port
        self.username = user
        self.password = password
        self.complement = complement
        # self.get_connection().connect()

    def get_connection(self):
        return MongoConnection(
            db=self.db,
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            complement=self.complement,
        )

    def create_document(self, db_name: str, document_class, **kwargs) -> object:
        document = document_class(**kwargs)
        document.save()
        return document

    def get_document(self, db_name: str, tenant: str, document_class, **kwargs) -> object:
        document = document_class.objects(**kwargs, context__tenant=tenant).first()
        # document.to_proto()
        return document

    def update_document(self, db_name: str, document_class, id: str, **kwargs) -> object:
        document = document_class.objects(id=id).first()
        document_class.objects(id=document.id).first().update(**kwargs)
        document.reload()
        return document

    def update(self, document_instance, **kwargs):
        document_instance.update(**kwargs)
        document_instance.reload()
        return document_instance

    def delete(self, document_instance):
        document_instance.delete()
        return document_instance

    def delete_document(self, db_name: str, document_class, id: str) -> object:
        document = document_class.objects(id=id).first()
        document.delete()
        return document

    def list_documents(
        self,
        db_name: str,
        tenant: str,
        document_class,
        fields: list = None,
        filter: dict = None,
        group_by: str = None,
        paginated: dict = None,
        sort_by: list = None,
    ) -> tuple[list, int]:
        """
        Parameters:
        fields (list): Optional list of fields to retrieve from the documents.
        filter (dict): Optional dictionary containing filter criteria for the query.
        group_by (str): Optional field to group results by.
        paginated (dict): Optional dictionary containing pagination information.
        sort_by (list): Optional list of fields to sort results by.

        Returns:
        list: A list of documents matching the specified criteria.
        """
        # Filter documents based on criteria provided

        if filter:
            query_set = document_class.objects(context__tenant=tenant).filter(__raw__=filter)
        else:
            query_set = document_class.objects(context__tenant=tenant)

        # Only retrieve specified fields
        if fields:
            query_set = query_set.only(*fields)

        # Group results by specified field
        if group_by:
            query_set = query_set.group_by(group_by)

        # Paginate results based on criteria provided
        if paginated:
            page = int(paginated.get("page") or 1)
            per_page = int(paginated.get("per_page") or 10)
            start = (page - 1) * per_page
            end = start + per_page
            query_set = query_set[start:end]

        # Sort results based on criteria provided
        if sort_by:
            query_set = query_set.order_by(*sort_by)

        # Return list of documents matching the specified criteria and total count of documents
        return list(query_set), query_set.count()

    def delete_documents(self, db_name, document_class, **kwargs):
        # with self.get_connection() as cnn:
        document = document_class.objects(**kwargs).delete()
        return document

    def update_embeded_document(
        self,
        db_name: str,
        document_class,
        filters: dict,
        update: dict,
        many: bool = False,
    ) -> object:
        # with self.get_connection() as cnn:
        if many:
            document_class.objects(**filters).update(**update)
            document = document_class.objects(**filters)
        else:
            document_class.objects(**filters).update_one(**update)
            document = document_class.objects(**filters).first()
        return document

    def batch_upsert(self, document_isntance, data):
        """
        Batch upserts a list of records into the database.
        Actualiza por lotes una lista de registros en la base de datos.

        Args:
            document_isntance(mongoengine.Document): The MongoEngine model to upsert into.
            data (list): A list of dictionaries containing the records to upsert.
        """

        bulk_operations = [
            UpdateOne(
                {"external_id": obj["external_id"]},
                {"$set": obj | {"tenant": data["context"]["tenant"], "updated_by": data["context"]["user"]}},
                upsert=True,
            )
            for obj in data["models"]
        ]

        result = document_isntance.bulk_write(bulk_operations)
        return result


class MongoConnection(object):
    """A MongoConnection class that can dynamically connect to a MongoDB database with MongoEngine and close the connection after each query.

    Args:
        host (str): The hostname or IP address of the MongoDB server.
        port (int): The port number of the MongoDB server.
        username (str): The username for the MongoDB database.
        password (str): The password for the MongoDB database.
        database (str): The name of the MongoDB database.
    """

    def __init__(self, host, port, db, username, password, complement):
        self.host = f"mongodb://{host}:{port}/?{'&'.join([f'{k}={v}' for (k, v) in complement.items()])}"
        self.port = port
        self.username = username
        self.password = password
        self.db = db

    def connect(self):
        """Connects to the MongoDB database.

        Returns:
            A MongoEngine connection object.
        """
        self.connection = mongo.connect(
            db=self.db,
            username=self.username,
            password=self.password,
            host=self.host,
        )
        return self.connection

    def close(self):
        """Closes the connection to the MongoDB database."""
        # self.connection.close()
        mongo.disconnect()

    def __enter__(self):
        """Enters a context manager.

        Returns:
            A MongoConnection object.
        """
        self.connect()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exits a context manager."""
        self.close()


class PolishNotationToMongoDB:
    def __init__(self, expression):
        self.expression = expression
        self.operators_logical = {
            "and": "$and",
            "or": "$or",
            "nor": "$nor",
            "not": "$not",
        }
        self.operators_comparison = {
            "=": "$eq",
            ">": "$gt",
            "<": "$lt",
            ">=": "$gte",
            "<=": "$lte",
            "in": "$in",
            "nin": "$nin",
            "!=": "$ne",
            "!like": "$not",
            "like": "$regex",
        }

    def is_logical_operator(self, token):
        if not isinstance(token, str):
            return False
        return token in self.operators_logical

    def is_comparison_operator(self, token):
        if not isinstance(token, str):
            return False
        return token in self.operators_comparison

    def is_tuple(self, token):
        return isinstance(token, tuple) and len(token) == 3

    def convert(self):
        operand_stack = []
        operator_stack = []

        for token in reversed(self.expression):
            if self.is_logical_operator(token):
                operator_stack.append(token)
            elif self.is_comparison_operator(token):
                operator_stack.append(token)
            elif self.is_tuple(token):
                field, old_operator, value = token
                if old_operator in self.operators_comparison:
                    options = {}
                    if old_operator == "like":
                        options = {"$options": "i"}
                    elif old_operator == "!like":
                        options = {
                            self.operators_comparison[old_operator]: {
                                "$regex": value,
                                "$options": "i",
                            }
                        }
                    operand_stack.append({field: {self.operators_comparison[old_operator]: value} | options})
                else:
                    raise ValueError(f"Unexpected operator: {old_operator}")
            else:
                raise ValueError(f"Unexpected token: {token}")

        while operator_stack:
            operator = operator_stack.pop()
            if operator in self.operators_logical:
                operands = []
                for _ in range(2):
                    operands.append(operand_stack.pop())
                operand_stack.append({self.operators_logical[operator]: operands})
            else:
                raise ValueError(f"Unexpected operator: {operator}")

        return operand_stack.pop()


class DBUtil(object):
    @classmethod
    def db_prepared_statement(
        cls,
        id: str,
        fields: base_pb2.Fields,
        filter: base_pb2.Filter,
        paginated: base_pb2.Paginated,
        group_by: base_pb2.GroupBy,
        sort_by: base_pb2.SortBy,
    ) -> dict:
        prepared_statement = {}
        prepared_statement["paginated"] = {
            "page": paginated.offset,
            "per_page": paginated.limit or 10,
        }
        if (ft := filter.ListFields()) or id:
            expression = [("_id", "=", cls.generate_object_id(id))]
            if ft:
                str_filter = filter.filter.replace("true", "True").replace("false", "False").replace("__", ".")
                expression = ast.literal_eval(str_filter)
                # reemplace filter id by _id and convert to ObjectId
                for idx, exp in enumerate(expression):
                    if isinstance(exp, tuple) and len(exp) == 3 and exp[0] == "id":
                        if type(exp[2]) == list:
                            expression[idx] = (
                                "_id",
                                exp[1],
                                [cls.generate_object_id(x) for x in exp[2]],
                            )
                            continue
                        expression[idx] = (
                            "_id",
                            exp[1],
                            cls.generate_object_id(exp[2]),
                        )
            filter_custom = PolishNotationToMongoDB(expression=expression).convert()
            prepared_statement["filter"] = filter_custom
        if group_by:
            prepared_statement["group_by"] = [x.name_field for x in group_by]
        if sort_by.ListFields():
            prepared_statement["sort_by"] = [cls.db_trans_sort(sort_by)]
        if fields:
            prepared_statement["fields"] = fields.name_field
        return prepared_statement

    @classmethod
    def db_trans_sort(cls, sort_by: base_pb2.SortBy) -> str:
        if not sort_by.name_field:
            return None
        return f"{'-' if sort_by.type == sort_by.DESC else '+'}{sort_by.name_field}"

    @classmethod
    def generate_object_id(cls, id=None):
        try:
            return ObjectId(id)
        except:
            return ObjectId(None)
